/***************************************************************************
 *             __________               __   ___.
 *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
 *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
 *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
 *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
 *                     \/            \/     \/    \/            \/
 * $Id: rolo_load_app.c,v 1.0 8/4/02 8:02PM randywood Exp $
 *
 * Copyright (C) 2002 Randy D. Wood
 *
 * All files in this archive are subject to the GNU General Public License.
 * See the file COPYING in the source tree root for full license agreement.
 *
 * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
 * KIND, either express or implied.
 *
 ****************************************************************************/

#ifndef SIMULATOR
#include "lcd.h"
#include "kernel.h"
#include "sprintf.h"
#include "button.h"
#include "file.h"
extern void mpeg_stop(void);
extern void system_init(void);
extern void i2c_init(void);
extern void rolo_move_app(int);
/***************************************************************************
 *
 *  Name: rolo_load_app_err(char *string)
 *          Scroll "string" on 1st line
 *          Display "Press Buttn" on 2nd line
 *          Wait for a button to be pressed
 *
 *  This routine is used to display an error message and wait for the
 *  user to press any button.
 *
 *
 ***************************************************************************/
void rolo_load_app_err(char *text)
{
    lcd_clear_display();
	lcd_puts(0, 1, "Press Buttn");
	lcd_puts_scroll(0, 0, text);
	button_get(true);
    lcd_stop_scroll();
}
/***************************************************************************
 *
 *  Name: rolo_load_app(char *filename,int scrambled)
 *          Filename must be a fully defined filename including the path and extension
 *          Scrambled == 0 application file has archos 6 byte header but is not scrambled (new file type such as .rbx)
 *                    != 0 applicaiton file is standard archos.mod format
 *
 *          This routine:
 *              1. Opens the application file
 *              2. Reads the file length in the file header and verifies that it matches the file length
 *              3. Reads and saves the file checksum in the file header
 *              Scrambled 4.  Verifies that there is room enough in mp3buf to read and descramble application file
 *              Scrambled 4a. Reads the application file into mp3buf offset by the length of the file
 *              Scrambled 4b. Descrambles the file to the start of mp3buf
 *              Unscrambled 4.  Verifies that there is room enough if mp3buf to read in the application file
 *              Unscrambled 4a. Reads the application file into the beginning of mp3buf
 *              5. Computes the checksum of the application program in memory and verifies it against the checksum from the file header
 *              6. Computes the number of longs words in the applications (rounded up)
 *              7. Calls the application move routine.
 *
 ***************************************************************************/
int rolo_load_app(char *filename,int scrambled)
{
    int fd,slen;
    unsigned long length,file_length,i;
    extern unsigned char mp3buf[],mp3end;
    unsigned short checksum,file_checksum;
/***************************************************************************
 *
 *  Stop mpeg stream, let things settle out, and display ROLO Loading
 *
 ***************************************************************************/
    mpeg_stop();                            /* stop mpeg stream                                             */
    sleep(HZ);                              /* give 1 second to stop                                        */
    lcd_clear_display();
    lcd_puts(0, 0, "ROLO...");
	lcd_puts(0, 1, "Loading");
    checksum=0;
    file_checksum=0;
/***************************************************************************
 * 
 * Open file and verify that open did not fail
 *
 ***************************************************************************/
    fd = open(filename, O_RDONLY);
    if(-1 == fd) {
        rolo_load_app_err("File not found");
	    return -1;
    }
/***************************************************************************
 *
 *  Read file length from header and compare to real file length
 *
 ***************************************************************************/
    length=lseek(fd,0,SEEK_END)-6;
    lseek(fd, 0, SEEK_SET);
    if(read(fd, &file_length, 4) != 4) {
        rolo_load_app_err("Error Reading File Length");
	    return -1;
    }
    if (length != file_length) {
        rolo_load_app_err("File length mismatch");
	    return -1;
    }
/***************************************************************************
*
*   Read and save checksum
*
 ***************************************************************************/
    lseek(fd, 4, SEEK_SET);
    if (read(fd, &file_checksum, 2) != 2) {
        rolo_load_app_err("Error Reading checksum");
	    return -1;
    }
    lseek(fd, 6, SEEK_SET);
/***************************************************************************
*
*   If Scrambled -  verify that file can be read and descrambled,
*                   read and descramble file
*
 ***************************************************************************/
    if (scrambled) {
        if ((&mp3buf[0] + (2*length)+4) >= &mp3end) {
            rolo_load_app_err("Not enough room to load file");
	        return -1;
        }
        if (read(fd, &mp3buf[length], length) != (int)length) {
            rolo_load_app_err("Error Reading File");
	        return -1;
        }
        /* descramble Thanks Bjorn  */
        slen = length/4;
        for (i = 0; i < length; i++) {
            unsigned long addr = ((i % slen) << 2) + i/slen;
            unsigned char data = mp3buf[i+length];
            data = ~((data >> 1) | ((data << 7) & 0x80)); /* poor man's ROR                                     */
            mp3buf[addr] = data;
        }
    }
    else {
/***************************************************************************
*
*   If not Scrambled - Read file
*
 ***************************************************************************/
        if ((&mp3buf[0] + length + 4) >= &mp3end) {
            rolo_load_app_err("Not enough room to load file");
	        return -1;
        }
        if (read(fd, &mp3buf[0], length) != (int)length) {
            rolo_load_app_err("Error Reading File");
	        return -1;
        }
    }
/***************************************************************************
 * 
 *  Compute checksum and verify against checksum from file header
 *
 ***************************************************************************/
    /* calculate checksum */
    for (i=0;i<length;i++)
        checksum += mp3buf[i];

    lcd_clear_display();
    if (checksum != file_checksum) {
        rolo_load_app_err("Checksum Error");
	    return -1;
    }
	lcd_puts(0, 1, "Moving   ");
/***************************************************************************
 *  Calling these 2 initialization routines was necessary to the the origional
 *  Archos version of the firmware to load and execute.
 ***************************************************************************/
    system_init();                              /* Initialize system for restart                            */
    i2c_init();                                 /* Init i2c bus - it seems like a good idea                 */
/***************************************************************************
 *
 *  Compute number of long words and call application move routine
 *
 ***************************************************************************/
    slen=(length+3)/4;                          /* Round up to next long word                               */
    rolo_move_app(slen);                        /* Call application mover                                   */
    return 0;
}
#else
void rolo_load_app_err(char *text)
{
    return;
}

#endif /* SIMULATOR */
